@@ -3,15 +3,17 @@ package MOP4Import::Declare;
 use 5.010;
 use strict;
 use warnings qw(FATAL all NONFATAL misc);
-our $VERSION = '0.049_002';
+our $VERSION = '0.051_001';
 use Carp;
 use mro qw/c3/;
 
 use constant DEBUG => $ENV{DEBUG_MOP4IMPORT};
 
-print STDERR "Using MOP4Import::VERSION = $VERSION (file '"
+use Sub::Util ();
+
+print STDERR "\nUsing ".__PACKAGE__. " = $VERSION (file '"
   . __FILE__ . "')\n"
-  if DEBUG and DEBUG >= 2;
+  if DEBUG;
 
 use MOP4Import::Opts
   qw/
@@ -21,10 +23,13 @@ use MOP4Import::Opts
     /;
 use MOP4Import::Util;
 use MOP4Import::FieldSpec;
+use MOP4Import::NamedCodeAttributes qw(MODIFY_CODE_ATTRIBUTES /^m4i_CODE_ATTR_/);
+
+#========================================
 
 our %FIELDS;
 
-sub import {
+sub import :MetaOnly {
   my ($myPack, @decls) = @_;
 
   m4i_log_start() if DEBUG;
@@ -54,14 +59,15 @@ sub always_exports {
   qw(-strict);
 }
 
-sub dispatch_declare {
+sub dispatch_declare :MetaOnly {
   (my $myPack, my Opts $opts, my (@decls)) = m4i_args(@_);
 
   foreach my $declSpec (@decls) {
 
     croak "Undefined pragma!" unless defined $declSpec;
 
-    if (not ref $declSpec) {
+    if (not ref $declSpec
+        or ref $declSpec eq 'Regexp') {
 
       $myPack->dispatch_import($opts, $declSpec);
 
@@ -85,30 +91,53 @@ our %SIGIL_MAP = qw(* GLOB
 		    @ ARRAY
 		    & CODE);
 
-sub dispatch_import {
+sub dispatch_import :MetaOnly {
   (my $myPack, my Opts $opts, my ($declSpec)) = m4i_args(@_);
 
   my ($name, $exported);
 
-  if ($declSpec =~ /^-([A-Za-z]\w*)$/) {
+  if (not ref $declSpec and $declSpec =~ /^-([A-Za-z]\w*)$/) {
 
     return $myPack->dispatch_declare_pragma($opts, $1);
 
-  } elsif ($declSpec =~ /^([\*\$\%\@\&])?([A-Za-z]\w*)$/) {
-
-    if ($1) {
-      my $kind = $SIGIL_MAP{$1};
-      $myPack->can("import_$kind")
-	->($myPack, $opts, $1, $kind, $2);
-    } else {
-      $myPack->import_NAME($opts => $2);
-    }
   } else {
-    croak "Invalid import spec: $declSpec";
+
+    $myPack->dispatch_import_no_pragma($opts, $declSpec);
   }
 }
 
-sub import_NAME {
+sub dispatch_import_no_pragma :MetaOnly {
+  (my $myPack, my Opts $opts, my (@declSpec)) = m4i_args(@_);
+  foreach my $declSpec (@declSpec) {
+
+    if (ref $declSpec eq 'Regexp') {
+      $myPack->import_by_regexp($opts, $declSpec);
+    }
+    elsif ($declSpec =~ /^([\*\$\%\@\&])?([A-Za-z]\w*)$/) {
+      if ($1) {
+        my $kind = $SIGIL_MAP{$1};
+        $myPack->can("import_$kind")
+          ->($myPack, $opts, $1, $kind, $2);
+      } else {
+        $myPack->import_NAME($opts => $2);
+      }
+    } else {
+      croak "Invalid import spec: $declSpec";
+    }
+  }
+}
+
+sub import_by_regexp :MetaOnly {
+  (my $myPack, my Opts $opts, my ($pattern)) = m4i_args(@_);
+
+  my $symtab = MOP4Import::Util::symtab($myPack);
+  foreach my $name (keys %$symtab) {
+    next unless $name =~ $pattern;
+    *{globref($opts->{destpkg}, $name)} = $symtab->{$name};
+  }
+}
+
+sub import_NAME :MetaOnly {
   (my $myPack, my Opts $opts, my ($name)) = m4i_args(@_);
 
   my $exported = safe_globref($myPack, $name);
@@ -119,7 +148,7 @@ sub import_NAME {
   *{globref($opts->{destpkg}, $name)} = $exported;
 }
 
-sub import_GLOB {
+sub import_GLOB :MetaOnly {
   (my $myPack, my Opts $opts, my ($sigil, $kind, $name)) = m4i_args(@_);
 
   my $exported = safe_globref($myPack, $name);
@@ -130,7 +159,7 @@ sub import_GLOB {
   *{globref($opts->{destpkg}, $name)} = $exported;
 }
 
-sub import_SIGIL {
+sub import_SIGIL :MetaOnly {
   (my $myPack, my Opts $opts, my ($sigil, $kind, $name)) = m4i_args(@_);
 
   my $exported = *{safe_globref($myPack, $name)}{$kind};
@@ -147,40 +176,57 @@ sub import_SIGIL {
 *import_HASH = *import_SIGIL; *import_HASH = *import_SIGIL;
 *import_CODE = *import_SIGIL; *import_CODE = *import_SIGIL;
 
-sub dispatch_declare_pragma {
+sub dispatch_declare_pragma :MetaOnly {
   (my $myPack, my Opts $opts, my ($pragma, @args)) = m4i_args(@_);
   if ($pragma =~ /^[A-Za-z]/
       and my $sub = $myPack->can("declare_$pragma")) {
     $sub->($myPack, $opts, @args);
   } else {
-    croak "Unknown pragma ".terse_dump($pragma)." in $opts->{destpkg}"
+    croak "No such pragma: \`use $myPack\ [".terse_dump($pragma)."]`"
       . $myPack->file_line_of($opts);
   }
 }
 
+#========================================
+
 # You may want to override these pragrams.
-sub declare_default_pragma {
+sub declare_default_pragma :MetaOnly {
   (my $myPack, my Opts $opts) = m4i_args(@_);
   $myPack->declare_c3($opts);
 }
 
-sub declare_strict {
+sub declare_strict :MetaOnly {
   (my $myPack, my Opts $opts) = m4i_args(@_);
   $_->import for qw(strict warnings); # I prefer fatalized warnings, but...
 }
 
 # Not enabled by default.
-sub declare_fatal {
+sub declare_fatal :MetaOnly {
   (my $myPack, my Opts $opts) = m4i_args(@_);
   warnings->import(qw(FATAL all NONFATAL misc));
 }
 
-sub declare_c3 {
+sub declare_c3 :MetaOnly {
   (my $myPack, my Opts $opts) = m4i_args(@_);
   mro::set_mro($opts->{destpkg}, 'c3');
 }
 
-sub declare_base {
+#========================================
+
+#
+# Just for readability
+#
+#   use XXX [import => qw/X Y Z/];
+#
+sub declare_import :MetaOnly {
+  (my $myPack, my Opts $opts, my (@import)) = m4i_args(@_);
+
+  $myPack->dispatch_import_no_pragma($opts, @import);
+}
+
+#========================================
+
+sub declare_fileless_base :MetaOnly {
   (my $myPack, my Opts $opts, my (@base)) = m4i_args(@_);
 
   $myPack->declare___add_isa($opts->{objpkg}, @base);
@@ -188,7 +234,9 @@ sub declare_base {
   $myPack->declare_fields($opts);
 }
 
-sub declare_parent {
+*declare_base = *declare_parent; *declare_base = *declare_parent;
+
+sub declare_parent :MetaOnly {
   (my $myPack, my Opts $opts, my (@base)) = m4i_args(@_);
 
   foreach my $fn (@base) {
@@ -196,10 +244,10 @@ sub declare_parent {
     require "$cp.pm";
   }
 
-  $myPack->declare_base($opts, @base);
+  $myPack->declare_fileless_base($opts, @base);
 }
 
-sub declare_as_base {
+sub declare_as_base :MetaOnly {
   (my $myPack, my Opts $opts, my (@fields)) = m4i_args(@_);
 
   print STDERR "Class $opts->{objpkg} inherits $myPack\n"
@@ -211,10 +259,10 @@ sub declare_as_base {
 
   $myPack->declare_fields($opts, @fields);
 
-  $myPack->declare_constant($opts, MY => $opts->{objpkg}, or_ignore => 1);
+  $myPack->declare_private_constant($opts, MY => $opts->{objpkg}, or_ignore => 1);
 }
 
-sub declare___add_isa {
+sub declare___add_isa :MetaOnly {
   my ($myPack, $objpkg, @parents) = @_;
 
   print STDERR "Class $objpkg extends ".terse_dump(@parents)."\n"
@@ -262,8 +310,16 @@ sub declare___add_isa {
   }
 }
 
-# XXX: previously was [as].
-sub declare_naming {
+# I'm afraid this 'as' pragma could invite ambiguous interpretation.
+# But in following case, I can't find any other pragma name.
+#
+#   use XXX [as => 'YYY']
+#
+# So, let's define `declare_as` as an alias of `declare_naming`.
+#
+*declare_as = *declare_naming; *declare_as = *declare_naming;
+
+sub declare_naming :MetaOnly {
   (my $myPack, my Opts $opts, my ($name)) = m4i_args(@_);
 
   unless (defined $name and $name ne '') {
@@ -273,26 +329,44 @@ sub declare_naming {
   $myPack->declare_constant($opts, $name => $myPack);
 }
 
-sub declare_inc {
+sub declare_inc :MetaOnly {
   (my $myPack, my Opts $opts, my ($pkg)) = m4i_args(@_);
   $pkg //= $opts->{objpkg};
   $pkg =~ s{::}{/}g;
   $INC{$pkg . '.pm'} = 1;
 }
 
-sub declare_constant {
+sub declare_constant :MetaOnly {
   (my $myPack, my Opts $opts, my ($name, $value, %opts)) = m4i_args(@_);
 
+  $myPack->declare_private_constant($opts, $name, $value, %opts);
+
+  my $export = MOP4Import::Util::ensure_symbol_has_array(
+    globref($opts->{objpkg}, 'EXPORT')
+  );
+  push @$export, $name;
+  print STDERR " constant $name is added to default exports of "
+    . $opts->{objpkg}. "\n" if DEBUG;
+}
+
+sub declare_private_constant :MetaOnly {
+  (my $myPack, my Opts $opts, my ($name, $value, %opts)) = m4i_args(@_);
+
+  my $or_ignore = delete $opts{or_ignore};
+  if (keys %opts) {
+    Carp::croak("Unknown options: ". join ", ", sort keys(%opts));
+  }
+
   my $my_sym = globref($opts->{objpkg}, $name);
   if (*{$my_sym}{CODE}) {
-    return if $opts{or_ignore};
+    return if $or_ignore;
     croak "constant $opts->{objpkg}::$name is already defined";
   }
 
-  *$my_sym = sub () {$value};
+  MOP4Import::Util::define_constant($my_sym, $value);
 }
 
-sub declare_fields {
+sub declare_fields :MetaOnly {
   (my $myPack, my Opts $opts, my (@fields)) = m4i_args(@_);
 
   my $extended = fields_hash($opts->{objpkg});
@@ -308,23 +382,56 @@ sub declare_fields {
       next if defined $extended->{$name};
       print STDERR "  Field $opts->{objpkg}.$name is inherited "
 	. "from $super_class.\n" if DEBUG;
-      $extended->{$name} = $super->{$name}; # XXX: clone?
+      my $anyFieldSpec = $extended->{$name} = do {
+        my $origin = $super->{$name};
+         # XXX: mark origin in this clone?
+        ref $origin ? $origin->clone : $origin;
+      };
       push @$fields_array, $name;
+      if (ref $anyFieldSpec) {
+        my FieldSpec $fs = $anyFieldSpec;
+        if ($fs->{json_type}) {
+          $myPack->declare___field_with_json_type(
+            $opts, $fs, json_type => $fs->{json_type}
+          );
+        }
+      }
     }
   }
 
-  $myPack->declare___field($opts, ref $_ ? @$_ : $_) for @fields;
+  {
+    my %dup;
+    foreach my $spec (@fields) {
+      my ($name) = ref $spec ? @$spec : $spec;
+      if ($dup{$name}++) {
+        croak "Duplicate field decl! $name";
+      }
+    }
+  }
+
+
+  my $field_class = do {
+    # XXX: objpkg? destpkg?
+    if (my $sub = $opts->{destpkg}->can("FieldSpec")) {
+      $sub->();
+    } else {
+      $myPack->FieldSpec;
+    }
+  };
+
+  print STDERR "  FieldSpec is: $field_class\n" if DEBUG;
+
+  $myPack->declare___field($opts, $field_class, ref $_ ? @$_ : $_) for @fields;
 
   $opts->{objpkg}; # XXX:
 }
 
-sub declare___field {
-  (my $myPack, my Opts $opts, my ($name, @rest)) = m4i_args(@_);
+sub declare___field :MetaOnly {
+  (my $myPack, my Opts $opts, my ($field_class, $name, @rest)) = m4i_args(@_);
   print STDERR "  Declaring field $opts->{objpkg}.$name " if DEBUG;
   my $extended = fields_hash($opts->{objpkg});
   my $fields_array = fields_array($opts->{objpkg});
 
-  my $field_class = $myPack->FieldSpec;
   my $spec = fields_hash($field_class);
   my (@early, @delayed);
   while (my ($k, $v) = splice @rest, 0, 2) {
@@ -346,13 +453,22 @@ sub declare___field {
 
   my FieldSpec $obj = $extended->{$name}
     = $field_class->new(@early, name => $name);
+  $obj->{package} = $opts->{objpkg};
   print STDERR " with $myPack $field_class => ", terse_dump($obj), "\n"
     if DEBUG;
   push @$fields_array, $name;
 
   # Create accessor for all public fields.
   if ($name =~ /^[a-z]/i and not $obj->{no_getter}) {
-    *{globref($opts->{objpkg}, $name)} = sub { $_[0]->{$name} };
+    if (my $sym = MOP4Import::Util::symtab($opts->{objpkg})->{$name}) {
+      if (*{$sym}{CODE}) {
+        croak "Accessor $opts->{objpkg}::$name is redefined!\nIf you really want to define the accessor by hand, please specify fields spec like: [$name => no_getter => 1, ...].";
+      }
+    }
+
+    *{globref($opts->{objpkg}, $name)}
+      = Sub::Util::set_subname(join("::", $opts->{objpkg}, $name)
+                               , sub :method { $_[0]->{$name} });
   }
 
   foreach my $delayed (@delayed) {
@@ -363,7 +479,7 @@ sub declare___field {
   $obj;
 }
 
-sub declare___field_with_default {
+sub declare___field_with_default :MetaOnly {
   (my $myPack, my Opts $opts, my FieldSpec $fs, my ($k, $v)) = m4i_args(@_);
 
   $fs->{$k} = $v;
@@ -371,11 +487,28 @@ sub declare___field_with_default {
   if (ref $v eq 'CODE') {
     *{globref($opts->{objpkg}, "default_$fs->{name}")} = $v;
   } else {
-    $myPack->declare_constant($opts, "default_$fs->{name}", $v);
+    $myPack->declare_private_constant($opts, "default_$fs->{name}", $v);
   }
 }
 
-sub declare_alias {
+sub JSON_TYPE_HANDLER {
+  require MOP4Import::Util::JSON_TYPE;
+  'MOP4Import::Util::JSON_TYPE';
+}
+
+sub declare___field_with_json_type :MetaOnly {
+  (my $myPack, my Opts $opts, my FieldSpec $fs, my ($kw, $typeSpec)) = m4i_args(@_);
+
+  print STDERR "  $fs->{name} json_type: spec=", terse_dump($typeSpec), "\n" if DEBUG;
+
+  my $json_type = $myPack->JSON_TYPE_HANDLER
+    ->register_json_type_of_field($opts->{objpkg}, $fs->{name}, $typeSpec);
+  $fs->{json_type} = $json_type;
+
+  #" result=", terse_dump($json_type), "\n"
+}
+
+sub declare_alias :MetaOnly {
   (my $myPack, my Opts $opts, my ($name, $alias)) = m4i_args(@_);
   print STDERR " Declaring alias $name in $opts->{destpkg} as $alias\n" if DEBUG;
   my $sym = globref($opts->{destpkg}, $name);
@@ -385,7 +518,27 @@ sub declare_alias {
   *$sym = sub () {$alias};
 }
 
-sub declare_map_methods {
+# Set(override) default value for inherited fields
+#
+# [defaults =>
+#    fieldName => defaultValue, ...
+# ]
+#
+sub declare_defaults :MetaOnly {
+  (my $myPack, my Opts $opts, my (@kvlist)) = m4i_args(@_);
+
+  my $fields = fields_hash($opts->{objpkg});
+
+  while (my ($k, $v) = splice @kvlist, 0, 2) {
+    my FieldSpec $fs = $fields->{$k}
+      or Carp::croak "No such field: $k";
+    $fs->{default} = $v;
+    my $fn = "default_$k";
+    $myPack->declare_private_constant($opts, $fn, $v);
+  }
+}
+
+sub declare_map_methods :MetaOnly {
   (my $myPack, my Opts $opts, my (@pairs)) = m4i_args(@_);
 
   foreach my $pair (@pairs) {
@@ -396,6 +549,32 @@ sub declare_map_methods {
   }
 }
 
+sub declare_carp_not :MetaOnly {
+  (my $myPack, my Opts $opts, my (@carp_not)) = m4i_args(@_);
+
+  unless (@carp_not) {
+    push @carp_not, $myPack;
+  }
+
+  my $name = 'CARP_NOT';
+
+  print STDERR "Declaring \@$opts->{objpkg}.$name = ".terse_dump(@carp_not)
+    if DEBUG;
+
+  *{globref($opts->{objpkg}, $name)} = \@carp_not;
+}
+
+{
+  #
+  # Below does equiv of `our @CARP_NOT = qw/ MOP4Import::Util /;`
+  #
+  __PACKAGE__->declare_carp_not(MOP4Import::Opts::m4i_fake(__PACKAGE__),
+                                qw/
+                                   MOP4Import::Util
+                                   /
+                                 );
+}
+
 1;
 
 
